
class ECConfig:

    class _ECConfig:
        CONFIG_PARAMS = {
            "ec_base_url": "https://envirocar.org/api/stable/",
            "ec_username": "",
            "ec_password": "",
            "number_of_processes": 4
        }

        def __init__(self):
            self.load_config()

        def load_config(self):
            for key, value in self.CONFIG_PARAMS.items():
                setattr(self, key, value)

    _instance = None

    def __init__(self):
        if not ECConfig._instance:
            ECConfig._instance = self._ECConfig()
        for item in ECConfig._instance.CONFIG_PARAMS:
            setattr(self, item, getattr(ECConfig._instance, item))
        


class BboxSelector:
    """ Class for creating bounding box requests """

    def __init__(self, bbox: [float]):
        self.min_x = bbox[0]
        self.min_y = bbox[1]
        self.max_x = bbox[2]
        self.max_y = bbox[3]

    @property
    def lower_left(self):
        return self.min_x, self.min_y

    @property
    def upper_right(self):
        return self.max_x, self.max_y

    @property
    def param(self):
        return { 'bbox': f'{self.min_x},{self.min_y},{self.max_x},{self.max_y}' }

class TimeSelector:
    def __init__(self, start_time, end_time):
        self.start_time = start_time
        self.end_time = end_time

    @property
    def param(self):
        return { 'after': f'{self.start_time}' } # TODO

class RequestParam:
    def __init__(self, path: str, method="GET", headers=None, params=None):
        self.path = path
        self.method = method
        self.headers = headers or []
        self.params = params or []


import json
import pandas as pd
import geopandas as gpd

from ..request_param import RequestParam, BboxSelector, TimeSelector
from ..download_client import DownloadClient

class TrackAPI():
    """Handles the API access to the enviroCar backend and returns queried results as dataframes"""
    TRACKS_ENDPOINT = "tracks"
    TRACK_ENDPOINT = "tracks/{}"
    USERTRACKS_ENDPOINT = "users/{}/tracks"
    
    def __init__(self, api_client=None):
        self.api_client = api_client or DownloadClient()

    def get_tracks(self, username=None, bbox:BboxSelector=None, time_interval:TimeSelector=None, num_results=10, page_limit=100):
        """Handles queries against the enviroCar api
        
        Keyword Arguments:
            username {str} -- the username to request tracks for (default: {None})
            bbox {BboxSelector} -- bbox query parameter (default: {None})
            time_interval {TimeSelector} -- time interval query parameter (default: {None})
            num_results {int} -- maximum number of tracks to request (default: {10})
            page_limit {int} -- page limit (default: {100})
        
        Returns:
            GeoDataFrame -- A GeoDataFrame containing the tracks matching the request parameters
        """
        path = self._get_path(username=username)

        # creating download_requests
        download_requests = []
        current_results = 0
        current_page = 1
        while current_results < num_results:
            request_params = {
                'limit': page_limit,
                'page': current_page
            }
            if bbox:
                request_params.update(bbox.param)

            request = RequestParam(path=path, params=request_params)
            download_requests.append(request)
            
            current_results += page_limit
            current_page += 1

        # request for /tracks
        tracks_meta_df = self.api_client.download(download_requests, decoder=_parse_tracks_list_df)
        tracks_meta_df = tracks_meta_df[:num_results]

        if not tracks_meta_df.empty: 
            ids = tracks_meta_df['track.id'].values
            tracks_df = self._get_tracks_by_ids(ids)
            return tracks_df

        return gpd.GeoDataFrame()

    def get_track(self, track_id: str):
        return self.api_client.download(
            RequestParam(path=self._get_path(trackid=track_id)), 
            decoder=_parse_track_df)

    def _get_tracks_by_ids(self, ids: [str]):
        download_requests = [RequestParam(path=self._get_path(trackid=id)) for id in ids]
        return self.api_client.download(download_requests, decoder=_parse_track_df)

    def _get_path(self, *, username=None, trackid=None):
        if username is None and trackid is None:
            return self.TRACKS_ENDPOINT
        if username:
            return self.USERTRACKS_ENDPOINT.format(username)
        if trackid:
            return self.TRACK_ENDPOINT.format(trackid)

def _parse_tracks_list_df(tracks_jsons):
    if not isinstance(tracks_jsons, list):
        tracks_jsons = [tracks_jsons]

    tracks_meta_df = pd.DataFrame()
    for tracks_json in tracks_jsons:
        if tracks_json:
            ec_data = json.loads(tracks_json)
            df = pd.json_normalize(ec_data, 'tracks')
            df.rename(columns=__rename_track_columns, inplace=True)
            tracks_meta_df = tracks_meta_df.append(df)

    return tracks_meta_df

def _parse_track_df(track_jsons):
    if not isinstance(track_jsons, list):
        track_jsons = [track_jsons]
    
    tracks_df = gpd.GeoDataFrame()
    for track_json in track_jsons:
        # read properties
        car_df = pd.json_normalize(json.loads(track_json)['properties'])
        car_df.columns = car_df.columns.str.replace('sensor.properties.', 'sensor.')
        car_df.rename(columns=__rename_track_columns, inplace=True)

        # read geojson values
        track_df = gpd.read_file(track_json)
        track_df = track_df.join(pd.json_normalize(track_df['phenomenons'])).drop(['phenomenons'], axis=1)
        
        # combine dataframes
        car_df = pd.concat([car_df]*len(track_df.index), ignore_index=True)
        tracks_df = tracks_df.append(track_df.join(car_df))
        
    return tracks_df

def __rename_track_columns(x):
    if not x.startswith('sensor'):
        return 'track.' + x
    return x


import requests

from .request_param import RequestParam
from ..exceptions import HttpFailedException, NotAuthorizedException


def handle_error_status(http_download):
    """Decorating function for handling enviroCar HTTP exceptions
    
    Arguments:
        http_download {function} -- function to wrap
    
    Raises:
        HttpFailedException: wrapper for client exceptions
        exception: other exceptions
    
    Returns:
        function -- decorated function
    """
    def decorate(self, request: RequestParam):
        try:
            return http_download(self, request)
        except requests.HTTPError as exception:
            if exception.response.status_code < requests.codes['internal_server_error']:
                if exception.response.status_code == requests.codes['unauthorized']:
                    message = "Unauthorized to access: {}".format(request.path)
                    raise NotAuthorizedException(message)
                else: 
                    message = "Failed to download from: {}".format(request.path)
                    raise HttpFailedException(message) from exception
            raise exception from exception
    return decorate



import logging
import requests 
from requests.auth import HTTPBasicAuth

import warnings
import concurrent.futures
import json
from urllib.parse import urljoin

from .client_config import ECConfig
from .request_param import RequestParam
from .utils import handle_error_status
from ..exceptions import HttpFailedException

LOG = logging.getLogger(__name__)

class DownloadClient:

    def __init__(self, *, config=None):
        self.config = config or ECConfig()

    def download(self, download_requests, decoder=None):
        if (isinstance(download_requests, RequestParam)):
            download_requests = [download_requests]

        with concurrent.futures.ThreadPoolExecutor(max_workers=self.config.number_of_processes) as executor:
            download_list = [executor.submit(self._download, request) for request in download_requests]

        result_list = []
        for future in download_list:
            try:
                decoded_data = future.result().decode('utf-8')
                result_list.append(decoded_data)
            except HttpFailedException as e:
                warnings.warn(str(e))
                result_list.append(None)

        if decoder:
            return decoder(result_list)
        return result_list

    @handle_error_status
    def _download(self, download_request: RequestParam):
        url = urljoin(self.config.ec_base_url, download_request.path)
        
        # set BasicAuth parameters
        auth = None
        if self.config.ec_username and self.config.ec_password:
            auth = HTTPBasicAuth(self.config.ec_username, self.config.ec_password)

        response = requests.request(
            download_request.method,
            url= url,
            auth=auth,
            headers=download_request.headers,
            params=download_request.params
        )

        response.raise_for_status()
        LOG.info("Successfully downloaded %s", url)
        return response.content



import io

from pip._internal.req import parse_requirements
from setuptools import setup, find_packages

# loading requirements
requirements = list(parse_requirements('requirements.txt', session='hack'))
requirements = [r.requirement for r in requirements]


def parse_long_description():
    return io.open('README.md', encoding="utf-8").read()

setup(
    name="envirocar-py",
    packages=find_packages(exclude=["tests", "tests.*"]),
    package_data={
        "": ["*.txt"]
    },
    include_package_data=True,
    version="0.0.3",
    description="Python Utilities for enviroCar",
    long_description=parse_long_description(),
    long_description_content_type="text/markdown",
    license="MIT",
    url="https://github.com/enviroCar/envirocar-py",
    keywords=["enviroCar", "trajectory", "xFCD"],
    install_requires=requirements,
    test_suite="tests",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ]
)



class HttpFailedException(Exception):
    """ General exception raised whenever communication failes """

class NotFoundException(HttpFailedException):
    """ Exception raised whenever a certain resource has not been found """

class NotAuthorizedException(HttpFailedException):
    """ Exception raised whenever the user doesn't has the rights to access a certain resource """

class MailNotConfirmedException(HttpFailedException):
    """ Exception raised whenever the user hasn't confirmed his mail """

